; SOCZ80 CP/M 3 BIOS (unbanked)
; 2014-05-24 Alan Cox
; based upon
; SOCZ80 CP/M 2.2 BIOS (or at least an attempt!)
; 2013-11-03 Will Sowerbutts
; SOCZ80 MP/M 2.1 XIOS
; 2013-11-19 Will Sowerbutts
; use zmac to assemble this
;
; zmac --rel cbios3.asm
; zmac --rel scb.asm
;
; then on CP/M
; link bnkbios3[os]=cbios3,scb
; gencpm
; Set up CP/M as banked and with 1 memory segment 01 6F 00
; With no video support you can use 01 7F 00 
;
; 0x7F leaves leaves the upper 16K of the bank for CP/M and bios image (you can
; go higher to about 0x8E - check the resulting image size). Note that the low
; base needs to be 0100 to allow for the IRQ vectors in bank 0
;
; To do
; - put CP/M 3 on the boot track
; - SD card support
; - switch for 'boot using display' (equivalent of the uart0/1 swap)
; - baud rate setting
; - support xmove block to block moves and build CP/M with an extra bank for
; - block caching
; - traditional CP/M would cache the CCP in a gap in bank 0 or 2. we don't
;   however seem to need to do that given the fixed A:
; - IRQ driven serial ?
; - TCP/IP
; - Finish switches to set boot device to include SD card once added
; - Switch to set fake 9600 baud delays ?
; - Third joystick ?
;
; Notes
; - the ram bank drives are described to CP/M as fixed drives. therefore
;   do not patch them via the monitor/downloader while within CP/M 
;

UART0_STATUS   		equ 0x00 ; [7: RX READY] [6: TX BUSY] [6 unused bits]
UART0_DATA     		equ 0x01

UART1_STATUS   		equ 0x08 ; [7: RX READY] [6: TX BUSY] [6 unused bits]
UART1_DATA     		equ 0x09

MMU_PAGESEL    		equ 0xF8
MMU_DATA		equ 0xFA
MMU_PERM       		equ 0xFB
MMU_FRAMEHI    		equ 0xFC ; In normal mode
MMU_FRAMELO    		equ 0xFD

MMU_PTR_VAL0		equ 0xFC ; In page 17 mode
MMU_PTR_VAL1		equ 0xFD
MMU_PTR_VAL2		equ 0xFE
MMU_PTR_VAL3		equ 0xFF

TIMER_STATUS            equ 0x10
TIMER_COMMAND           equ 0x11
TIMER_VAL0              equ 0x14
TIMER_VAL1              equ 0x15
TIMER_VAL2              equ 0x16
TIMER_VAL3              equ 0x17

TIMCMD_INTACK           equ 0x00
TIMCMD_UPRESET          equ 0x01
TIMCMD_UPLATCH          equ 0x02
TIMCMD_DOWNRESET        equ 0x03
TIMCMD_SEL_UPVAL        equ 0x10
TIMCMD_SEL_UPLATCH      equ 0x11
TIMCMD_SEL_DOWNVAL      equ 0x12
TIMCMD_SEL_DOWNRESET    equ 0x13

; SPI connected to the flash chip
SPI_CHIPSELECT          equ 0x18
SPI_STATUS              equ 0x19
SPI_TX                  equ 0x1a
SPI_RX                  equ 0x1b
SPI_DIVISOR             equ 0x1c

; SPI connected to the SD card
SD_CHIPSELECT           equ 0x30
SD_STATUS               equ 0x31
SD_TX                   equ 0x32
SD_RX                   equ 0x33
SD_DIVISOR              equ 0x34

; Papilio Pro flash type
FLASH_MAN               equ 0xc2
FLASH_DEV1              equ 0x20
FLASH_DEV2              equ 0x17

; Joystick
STICK0			equ 0xEC
STICK1			equ 0xED

; Switches
SWITCHES		equ 0x20


	.extrn SCBBASE
	.extrn CIVEC
	.extrn COVEC
	.extrn AIVEC
	.extrn AOVEC
	.extrn LOVEC
	.extrn BNKBF
	.extrn CRDMA
	.extrn CRDSK
	.extrn VINFO
	.extrn RESEL
	.extrn FX
	.extrn USRCD
	.extrn MLTIO
	.extrn ERMDE
	.extrn ERDSK
	.extrn MEDIA
	.extrn BFLGS
	.extrn DATE
	.extrn HOUR
	.extrn MIN
	.extrn SEC
	.extrn ERJMP
	.extrn MXTPA

		cseg

bdos_smulti    equ 44
bdos_open      equ 15
bdos_setdma    equ 26
bdos_read      equ 20
bdos	       equ 0x0005   ; BDOS jump vector
msize    equ 64 ; KB of system RAM
cdisk    equ 0x0004 ; current disk number, 0=a, 1=b ... 15=p apparently also current user number is in here (in the top 4 bits?)
iobyte   equ 0x0003 ; intel "standard" i/o byte

; jump vector table used by CP/M
            jp boot         ; cold start
wboote:     jp wboot        ; warm start
            jp const        ; console status
            jp conin        ; console character in
            jp conout       ; console character out
            jp list         ; list character out
            jp auxout       ; auxiliary character out
            jp auxin        ; auxiliary character in
            jp home         ; move disk head to home position
            jp seldsk       ; select disk
            jp seltrk       ; set track number
            jp setsec       ; set setor number
            jp setdma       ; set DMA address
            jp read         ; read disk
            jp write        ; write disk
            jp listst       ; return list status
            jp sectran      ; sector translate
            jp conost	    ; console output ready
            jp auxist	    ; aux input is ready
            jp auxost	    ; aux output is ready
            jp devtbl	    ; device table
            jp devini	    ; devince init
            jp drvtbl	    ; drive table
            jp multio	    ; multi block I/O
            jp flush	    ; flush data to disc
            jp move	    ; block move (including banking)
            jp time	    ; get current data and time into SCB
            jp selmem	    ; select memory bank. Special: preserve regs
            jp setbnk	    ; set the bank for the next read/write sector
            jp xmove	    ; set up for cross bank move
            jp userf	    ; extended BIOS
            jp 0	    ; reserved
            jp 0	    ; reserved

; disk parameter header (16 bytes for each drive), see page 6-28 in CP/M 2.2 operating system manual
; CP/M 3 style disk parameter header
;
; We keep private data in front of the dph pointer
            dw write_rd
            dw read_rd
dpbase0:
            ; disk 0 (A)
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblk        ; DPB (disk parameter block)
            dw 0	    ; CSV (unique scratch pad used to check for changed disks)
            dw 0xFFFE       ; ALV (unique scratch pad for allocation information)
	    dw 0xFFFE	    ; Directory buffer control block (auto assign)
	    dw 0xFFFF       ; Data BCB (none - 128 byte sectors)
            dw 0xFFFE       ; HASH
            db 0x00         ; HASH bank
            ; end of disk 0

            dw write_rd
            dw read_rd
dpbase1:
            ; disk 1 (B)
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblk        ; DPB (disk parameter block)
            dw 0            ; CSV (unique scratch pad used to check for changed disks)
            dw 0xFFFE       ; ALV (unique scratch pad for allocation information)
	    dw 0xFFFE	    ; Directory buffer control block (auto assign)
	    dw 0xFFFF       ; Data BCB (none - 128 byte sectors)
            dw 0xFFFE       ; HASH
            db 0x00         ; HASH bank
            ; end of disk 1

            dw write_rd
            dw read_rd
dpbase2:
            ; disk 1 (C)
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblk        ; DPB (disk parameter block)
            dw 0            ; CSV (unique scratch pad used to check for changed disks)
            dw 0xFFFE       ; ALV (unique scratch pad for allocation information)
	    dw 0xFFFE	    ; Directory buffer control block (auto assign)
	    dw 0xFFFF       ; Data BCB (none - 128 byte sectors)
            dw 0xFFFE       ; HASH
            db 0x00         ; HASH bank
            ; end of disk 2

            dw no_write
            dw flash_read
dpbasef0:
            ; flash disk 0 (N)
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblk        ; DPB (disk parameter block)
            dw 0            ; CSV (unique scratch pad used to check for changed disks)
            dw 0xFFFE       ; ALV (unique scratch pad for allocation information)
	    dw 0xFFFE	    ; Directory buffer control block (auto assign)
	    dw 0xFFFF       ; Data BCB (none - 128 byte sectors)
            dw 0xFFFE       ; HASH
            db 0x00         ; HASH bank
            ; end of disk 1


            dw no_write
            dw flash_read
dpbasef1:
            ; flash disk 1 (O)
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblk        ; DPB (disk parameter block)
            dw 0            ; CSV (unique scratch pad used to check for changed disks)
            dw 0xFFFE       ; ALV (unique scratch pad for allocation information)
	    dw 0xFFFE	    ; Directory buffer control block (auto assign)
	    dw 0xFFFF       ; Data BCB (none - 128 byte sectors)
            dw 0xFFFE       ; HASH
            db 0x00         ; HASH bank
            ; end of disk 1

            dw no_write
            dw flash_read
dpbasef2:
            ; flash disk 2 (P)
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblk        ; DPB (disk parameter block)
            dw 0            ; CSV (unique scratch pad used to check for changed disks)
            dw 0xFFFE       ; ALV (unique scratch pad for allocation information)
	    dw 0xFFFE	    ; Directory buffer control block (auto assign)
	    dw 0xFFFF       ; Data BCB (none - 128 byte sectors)
            dw 0xFFFE       ; HASH
            db 0x00         ; HASH bank
            ; end of disk 1

dpbasesd:
            ; SD card, use the 512MB drive mappings that Z80Pack uses
            ; to keep life simple
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblksd      ; DPB (disk parameter block)
            dw 0            ; CSV (unique scratch pad used to check for changed disks)
            dw 0xFFFE       ; ALV (unique scratch pad for allocation information)
	    dw 0xFFFE	    ; Directory buffer control block (auto assign)
	    dw 0xFFFF       ; Data BCB (none - 128 byte sectors)
            dw 0xFFFE       ; HASH
            db 0x00         ; HASH bank
            ; end of disk 1

dphtab	    dw dpbase0		; A:
	    dw dpbase1		; B:
ram3onoff:  dw dpbase2          ; C:
	    dw 0

	    dw 0		; E:
            dw 0
            dw 0
	    dw 0

	    dw 0		; I:
            dw 0
            dw 0
	    dw 0

	    dw 0		; M:
dphflash:   dw dpbasef0		; N-P - flash drives (R/O)
            dw dpbasef1
flash3onoff:dw dpbasef2

ndisks      equ 16           ; number of disks we defined

; disk parameter block (can be shared by all disks with same configuration)
; my RAM disks are each 2MB in size;
;     track is 16384 bytes (128 x 128 sectors per track)
;     there are 128 tracks (128 x 128 x 128 = 2MB)
;     track 0 is reserved for the operating system, so there are 2048-16 = 2032 user sectors, and 16KB reserved for OS boot storage (which is plenty)
; using 1K sectors here resulted in a problems. don't do that! with 4K sectors it seems much happier.
dpblk:
            dw 128          ; SPT: number of 128 byte sectors per track
            db 5            ; BSH: block shift factor (see manual for table)
            db 31           ; BLM: block mask (see manual for table)
            db 1            ; EXM: extent mask (see manual for table, using entries marked N/A turns out to be a bad idea!)
            dw 508          ; DSM: (disk bytes / block bytes) - 1, change alv00/01 etc if you change this; this is the number of the last sector on the disk, excluding system tracks (ie more system tracks -> this gets smaller)
            dw 511          ; DRM: directory max entries - 1
            db 0xf0         ; AL0: directory sector allocation bitmask byte 0
            db 0x00         ; AL1: directory sector allocation bitmask byte 1
            dw 0x8000       ; CKS: check size (0x8000 = fixed media, no check)
            dw 1            ; OFF: track offset (number of system tracks)
	    db 0	    ; PSH - 128 byte sectors
	    db 0	    ; PHM - 128 byte sectors

;
;  512MB SD card
;
dpblksd:
            dw 16384        ; SPT: number of 128 byte sectors per track
            db 7            ; BSH: block shift factor (see manual for table)
            db 127          ; BLM: block mask (see manual for table)
            db 0            ; EXM: extent mask (see manual for table, using entries marked N/A turns out to be a bad idea!)
            dw 0x7FFF       ; DSM: (disk bytes / block bytes) - 1, change alv00/01 etc if you change this; this is the number of the last sector on the disk, excluding system tracks (ie more system tracks -> this gets smaller)
            dw 8191         ; DRM: directory max entries - 1
            db 0xff         ; AL0: directory sector allocation bitmask byte 0
            db 0xff         ; AL1: directory sector allocation bitmask byte 1
            dw 0x8000       ; CKS: check size (0x8000 = fixed media, no check)
            dw 0            ; OFF: track offset (number of system tracks)
	    db 2	    ; PSH - 512 byte sectors
	    db 3	    ; PHM - 512 byte sectors

; The device utility table.

MB_INPUT    equ		1	; We do input
MB_OUTPUT   equ		2	; We do output
MB_INOUT    equ		3	; We do both
MB_SOFTBAUD equ		4       ; Soft baud rates
MB_SERIAL   equ         8       ; Serial protocol
MB_XONXOFF  equ         16	; XON/XOF protocol

BAUD_NONE    equ	0
BAUD_50	     equ	1
BAUD_75	     equ	2
BAUD_110     equ	3
BAUD_134     equ	4	; Actuall 134.5
BAUD_150     equ	5
BAUD_300     equ	6
BAUD_600     equ	7
BAUD_1200    equ	8
BAUD_1800    equ	9
BAUD_2400    equ	10
BAUD_3600    equ	11
BAUD_4800    equ	12
BAUD_7200    equ	13
BAUD_9600    equ	14
BAUD_19200   equ	15

;
;	For now and testing this will do fine
;
chrtbl:
	    db 'UART0 '			; device 0
	    db MB_INOUT + MB_SERIAL
	    db BAUD_9600
	    db 'UART1 '			; device 1
	    db MB_INOUT + MB_SERIAL
	    db BAUD_9600
	    db 'CRT   '			; device 2
	    db MB_INOUT
	    db BAUD_19200
	    db 'MOUSE '			; device 3
	    db MB_INPUT
	    db BAUD_9600
	    db 'JSTICK'			; device 4
	    db MB_INPUT
	    db BAUD_9600

	    db 0	    ; End of table

	    ds 128	    ; Provide a boot time stack
bootstack:
            ds 64           ; Console stack
constack:   dw 0            ; Saved stack we switched from

; bios functions follow
boot:       
            ; perform standard CPM initialisation
	    ld sp, bootstack
            xor a
            ld (iobyte), a
            ld (cdisk), a
            ld a, 128             ; UART 0
            ld (CIVEC+1), a
            ld (COVEC+1), a
            ; FIXME: check for a UART 1
            ld a, 64              ; UART 1
            ld (AIVEC+1), a
            ld (AOVEC+1), a
            ld a, 64
            ld (LOVEC+1), a

            ; say hello before we switch to bank 1
            ld hl, bootmsg
            call strout
	    ld a, 0xc3	    ; put the IRQ vector in both banks
	    ld (0x38), a
	    ld hl, clockint
	    ld (0x39), hl   ; write a copy into RST 38
	    im 1
	    ld a, 1
	    call selmem
	    ld a, 0xc3	    ; put the IRQ vector in both banks
	    ld (0x38), a
	    ld hl, clockint
	    ld (0x39), hl   ; write a copy into RST 38
	    xor a
	    call selmem	    ; back to bank 0 as clockinit is banked
	    call clockinit
            call coninit    ; also banked
            call spi_init    ; initialise the ROM disk SPI
            call switchinit ; switch features
            ei
            jp gocpm

warmboot    db 'Warm boot',13,10,0

wboot:    
	    ld sp, bootstack

gocpm:
	; Remember in banked mode we need to write these into bank 1 for
	; CCP to use
	    ld a, 1
	    call selmem
            ld a, 0xc3      ; 0xc3 is a jmp instruction
            ld (0), a       ; write JMP instruction at 0x0000
            ld (5), a       ; write JMP instruction at 0x0005
            ld hl, wboote
            ld (1), hl      ; write vector for warm boot at 0x0001, 0x0002.
            ld hl, (MXTPA)  ; BDOS entry point
            ld (6), hl      ; write vector for BDOS at 0x0006, 0x0007.

ccpload:
	    xor a	    ; set up our FCB
	    ld (ccpext), a
	    ld (ccprc), a
	    ld hl, 0
	    ld (fcbnr), hl
	    ld de, ccpfcb
	    ld c, bdos_open
	    call bdos
	    inc a
	    jr nz, openok   ; opened CCP.COM ok
	    ld hl, noccp
	    call strout	    ; 
	    call conin
	    jr ccpload
readfail:   ld hl, badccp
	    call strout	    ; 'try again'
	    call conin
	    jr ccpload

openok:
	    ld de, 0x100
	    ld c, bdos_setdma
	    call bdos
	    ld de, 128
	    ld c, bdos_smulti
	    call bdos
	    ld de, ccpfcb
	    ld c, bdos_read
	    call bdos

	    cmp 1
	    jr nz,readfail	    ; EOF is good
ccpok:
            ld bc, 0x0080   ; default DMA address
            call setdma     ; configure DMA
	    ld a, 1	    ; single sector I/O
	    ld (MLTIO), a   ; set default blocking
	    call selmem	    ; CCP is in bank 1
            ld a, (cdisk)   ; get current disk
            ld c, a         ; send to ccp
            jp 0x100          ; and we're off!

ccpfcb:	    db 1
	    db 'CCP     COM'
ccpext:	    db 0,0,0
ccprc:	    db 0
	    db 0,0,0,0,0,0,0,0
	    db 0,0,0,0,0,0,0,0
fcbnr:	    db 0,0,0
	    db 0,0,0

noccp:	    db 'CCP.COM not found',10,13,0
badccp:	    db 'CCP.COM read failed',10,13,0
;
;	CP/M 3 supports device redirection. All the console type ops become
;       the sum of the devices assigned.
;
;	This looks uglier but it actually makes life simple as all the
;	aux and reader stuff just becomes calls to the same code with
;	the right vector loaded
;
;
;	Supporting devices (here so JR reaches)
;
list:       ; write character to listing device (we don't have one!)
            ld a, (LOVEC+1)
            jp outwrite

listst:     ; return list device status
            ld a, (LOVEC+1)
            jp outstatus

auxout:      ; write character to the aux device
            ld a, (AOVEC+1)
            jp outwrite

auxin:      ; read character from the aux device
            ld a, (AIVEC+1)
            jr inread

auxist:	    ; auxiliary is ready for input
            ld a, (AIVEC+1)
            jr instatus

auxost:	    ; auxiliary is ready for output
            ld a, (AOVEC+1)
	    jp outstatus

;
;	Console and helpers
;
const:
            ld a, (CIVEC+1)

            ;
            ;	Implement all our input status methods with I/O
            ;   redirection. FIXME: spot physical lack of UART1
            ;
instatus:   ld l, a
            bit 7, l
            jr z, no_st_uart0
            in a, (UART0_STATUS)
            bit 7, a
            jr nz, const_ready
no_st_uart0:
            bit 6, l
            jr z, no_st_uart1
            in a, (UART1_STATUS)
            bit 7, a
            jr nz, const_ready
no_st_uart1:
            ;; FIXME: PS/2 ports
            bit 3, l 				; joystick
            jr z, const_notready
            in a, (STICK0)
            cp 255
            jr z, no_stick0
            or a
            jr nz, const_ready
no_stick0:  in a, (STICK1)
            cp 255
            jr z, const_notready
            or a
            jr nz, const_ready            
const_notready:
	    xor a
            ret
const_ready:
            ; we're ready
            ld a, 0xff
            ret
            
conin:      ; read character from console into A; wait if no character ready
            ld a, (CIVEC+1)
            ;
            ;	Implement all our input reading methods with I/O
            ;   redirection. FIXME: spot physical lack of UART1
            ;
            ld e, a
inread:     bit 7, e
            jr z, conin_u1
conin_try:  in a, (UART0_STATUS)
            bit 7, a
            jr z, conin_u1 ; keep waiting if no character ready
            in a, (UART0_DATA) ; read character
conin_fix:  ; fix backspace
            cp 0x7f ; backspace?
            ret nz
            ld a, 8 ; ctrl-h
            ret
conin_u1:   bit 6, e
            jr z, conin_joy
            in a, (UART1_STATUS)
            bit 7, a
            jr z, conin_joy ; keep waiting if no character ready
            in a, (UART1_DATA) ; read character
            jr conin_fix
conin_joy:  ; Slightly trickier as we need to read both stick evenly
            bit 3, e
            jr z, inread
            ld hl, sticktock
            ld a, 1
            xor (hl)
            ld (hl), a
            ld c, a
            in a, (c)
            cp 255              ; no stick
            jr z, inread
            or a                ; stick but no activity
            jr z, inread
;
; Fall through for now - rework when add PS/2
;
conin_dostick1:
            ld hl, stickmap1
            bit 0, c
            jr nz, conin_stickchar
conin_dostick:
            ld hl, stickmap0
conin_stickchar:
            ld c, a
            xor a
            bit 0, c
            jr nz, js_doy
            inc a
            bit 1, c
            jr z, js_doy
            inc a
js_doy:     bit 2, c
            jr nz, js_char
            add a, 3
            bit 3, c
            jr z, js_char
            add a,3
js_char:
            ld d, 0
            ld e, a
            add hl, de
            ld a, (hl)
            bit 4, c
            ret z
            res 5, a          ; upper case
            ret

sticktock:  db STICK0         ; cycle the read order so games see both
                              ; sticks in even flow
stickmap0:  db 'qweasdzxc'
stickmap1:  db 'rtyfghvbn'

conost:     ; check if console is ready for output (CPM 3 onwards)
            jr conost_nou1

            ld a, (COVEC+1)
            ;
            ;	Implement all our output status methods with I/O
            ;   redirection. FIXME: spot physical lack of UART1
            ;
outstatus:  ld l, a
conost_try: bit 7, l
            jr z, conost_nou0
	    in a, (UART0_STATUS);
	    and 0x40
	    jr z, conost_try
conost_nou0:bit 6, l
            jr z, conost_nou1
	    in a, (UART1_STATUS);
	    and 0x40
	    jr z, conost_try
conost_nou1:                  
            ;; CRT is always 'ready'
	    ld a,0xff
	    ret         

conout:     ; write chracter from C to console
            ld a, (COVEC+1)
            ;
            ;	Implement all our output methods with I/O
            ;   redirection. FIXME: spot physical lack of UART1
            ;
outwrite:   ld l, a
            bit 7, l
            jr z, conout_nou0
u0_wait:    in a, (UART0_STATUS)
            bit 6, a
            jr nz, u0_wait ; loop again if transmitter is busy
            ld a, c
            out (UART0_DATA), a ; transmit character
conout_nou0:
            bit 6, l
            jr z, conout_vt
u1_wait:    in a, (UART1_STATUS)
            bit 6, a
            jr nz, u0_wait ; loop again if transmitter is busy
            ld a, c
            out (UART0_DATA), a ; transmit character
conout_vt:  bit 5,l
            ret z
            ld a, (VID_MODE)   ; Graphics found ?
            or a
            ret z              ; Nope.
            call console       ; Banked interface call
            ret		       ; Optimise later

;
;	Rules for banked character I/O
;	- swap stacks to a local stack in common
;	- save the bank
;	- switch to bank 0
;	- invoke the I/O routine
;	- switch back
;	- restore the stack
;
console:	; Preserve C (and save the bank here too)
            ld a, (activebank)
            ld b, a
            ld hl, 0
            add hl, sp
            ld (constack), hl
            ld sp, constack    ; Switch stack
            push bc            ; Stack these onto our private stack
            xor a              ; Go to bank 0
            call selmem
	    call conentry      ; Console entry point
            pop bc
            ld a, b
            call selmem
            ld sp, (constack)
            ret


seldsk:     ; select disk indicated by register C
            ld a, c
	    ld hl,0
            cp ndisks
            ret nc  ; return (with error code) if C >= ndisks ie illegal drive
	    add a
	    ld l,a
	    ld de, dphtab    ; find this disk in dphtab
	    add hl, de
	    ld e,(hl)
	    inc hl
	    ld d,(hl)
	    ex de,hl
	    ld a, l	     ; if the entry is 0 return 0 in hl
	    or h
	    ret z
	    ld a, c	     ; if not also set curdisk
            ld (curdisk), a  ; store current disk
            ld (diskptr), hl ; store the disk pointer
            ret

home:       ld c, 0
            ; fall through into seltrk
seltrk:     ; set track given by register BC
            ld a, c
            ld (curtrack), a
            ret

setsec:     ; set sector given by register BC
	    ; for CP/M 3 this is a *physical* sector
            ld (cursector), bc
            ret

sectran:    ; logical to physical sector translation
            ; HL=BC ie 1:1 mapping (no translation)
            ; in CP/M 3 this is physical sectors but as we don't translate
	    ; we don't care
	    ; we will need to check this if we ever add SD support
            ld h, b
            ld l, c
            ret

devtbl:	    ; device table, 0 if not implemented
	    ; we don't really implement it yet but we will want it for
	    ; baud rate setting later
	    ld hl, chrtbl
	    ret

devini:	    ; re-initialize device C, called on baud/mode change
            ; nothing to do for now
            ret

drvtbl:	    ; drive table - return drive table in HL or 0
	    ; pointer to an array of 16 DPH blocks
	    ; does not *have* to be in common space (but usually is)
	    ; must end with ld hl, address (GENCPM snoops this!)
	    ld hl, dphtab
	    ret

multio:	    ; BDOS is going to transfer consective sectors
            ; basically a deblocker hint. We could use this to do one
	    ; MMU switch and bulk copy in theory. May be relevant if SD
	    ; support is added
	    xor a
            ret

flush:      ; Flush any pending data to disc
            ; FIXME: for now just report ok. If we support the SD then
            ; we may need to worry about these later
            xor a
            ret

move:       ; copy BC bytes from DE->HL. If move is active then
            ; do bank copies. Must support 128 byte banked copies, more optional
	    ; FIXME: bank support here using the MMU bypass and a common
	    ; buffer would make CP/M happier
;	    ld a,(xmove_sec) 	; 0xFF means none
;	    inc a
;	    jnz banked_copy
            ex de, hl
            ldir
            ex de,hl
            ; return de,hl as next bytes to copy 
            ret

xmove:      ; set the next move to copy between banks. C = source, B = dest
            ; CP/M weirdomatic warning - if unimplemented this must just be
            ; a ret. This is really meant for platforms that can do fast
            ; direct interbank copies - eg via a DMA engine
	    ret

;
;	    ld (xmove_sec), bc
;            ret
;
;mmu_aimload ; on entry a is the bank, de is the address. Our bus address
;	    ; is therefore hl + 48K * bank + 16K if not bank 0
;	    ; DE and BC are preserved to make life easier for the caller
;	    or a		;	 * 2
;	    jr nz,mmu_aim_high	;	In bank 0 (lowest 64K)
;	    ;
;	    ;	We special case bank 0 - its easier that way
; 	    ;
;	    ld a, 0xFF
;	    out (MMU_PAGESEL), a	; MMU special bank
;	    xor a
;	    out (MMU_PTR_VAL0), a	; VA = PA
;	    out (MMU_PTR_VAL1), a
;	    ld a, h
;	    out (MMU_PTR_VAL2), a
;	    ld a, l
;	    out (MMU_PTR_VAL3), a
;	    ret
;
;	    ; compute 16384 + 49152 * bank + address. This needs 24bit
;	    ; maths
;mmu_aim_high:
;	    rrca		;	 
;	    rrca		; 
;	    rrca		;	 bank * 0x20
;
;	    push de		;	preserve DE
;	    push hl		;	save the address
;
;				;	HL is used to compute the upper
;				;	two bytes of the 24bit value
;
;	    ld l, a		;	Bank number * 0x20
;	    ld h, 0		;	Need to do the last bits in 16bit
;	    add hl, hl		;	 * 0x40
;	    push hl
;	    add hl, hl		;	 * 0x80
;	    pop de
;	    add hl, de		;	 * 0xC0
;	    ld de, 0x0040	;	48 * base + 16K (for common)
;	    add hl, de		;	Add in the offset for the common
;	    ld a, 0xff
; 	    out (MMU_PAGESEL), a;	Magic mode
;	    xor a
;	    out (MMU_PTR_VAL0), a;	High 2 bits
;	    pop de		;	Retrieve the virtual address
;	    ld a, e
;	    out (MMU_PTR_VAL3), a;	Low byte is straight from virtual
;	    ld e, d		;	Now add the upper byte to HL
;	    ld d, 0
;	    add hl,de		;	Gives us the middle physical addr
;	    ld a, l		;	bytes
;	    out (MMU_PTR_VAL2), a
;	    ld a, h
;	    out (MMU_PTR_VAL1), a
;
;	    pop de		;	Restore the saved HL
;	    ret
;
;banked_copy:
;	    ; BC bytes from DE to HL
;	    push de		; save source addr
;	    push bc		; and length
;	    ld a, (xmove_sec)	; source bank
;	    ex de, hl		; source into HL, DE will be saved
;				; by the call
;	    call mmu_aimload	; MMU DMA now set up for address at A/src
;	    ld hl, xmove_buf	; HL is now free, load into the bounce buffer
;	    ld b,c		; count will be <= 128
;	    ld c, MMU_DATA	; data port
;	    push bc		; save for the write phase
;	    inir		; copy
;	    ld a, (xmove_sec+1) ; destination bank
;	    ex de,hl		; destination back into HL
;	    call mmu_aimload	; MMU DMA now set up for address at A/dst
;	    pop bc		; Quick way to get the count and port back
;	    ld hl, xmove_buf	; Holds the temporary copy we made
;	    otir		; bulk copy via the MMU DMA
;			        ; this leaves HL correct
;	    ld a, 0xFF
;	    ld (xmove_sec), a	; normal copy mode
;	    pop bc
;	    ex de,hl		; save destination
;	    pop hl		; recover source
;	    add hl, bc		; add length
;	    ex de,hl		; swap back the right way
;	    ret

selmem:     ; set the current bank to A, preserve all registers except A
            ; bank 0 is CP/M, bank 1 is TPA, bank 2+ are RAMdisc etc
	    ;
	    ; Note: we map bank 0 in an odd way. Our address map looks like
            ; this
            ; 0000-00FF: reserved (IRQ vector copies)
            ; 0100-6FFF: Given to CP/M for data
            ; 7000-7FFF: Video RAM (physical 2010000)
            ; 8000-BFFF: CP/M banked executable
            ; C000-FFFF: common

BANKED_PAGES equ 12   ; 0xc0 / 0x10 = 12 pages banked
	    push hl
	    ld hl, activebank
	    cmp (hl)
	    jz noswitch
	    ld (hl), a
	    push de
	    push bc
            add a     ; *2
            add a     ; *4
            ld l, a   ; store copy ie L=bank*4
            add a     ; *8
            add l     ; *12 (*4+*8)
            ld l, a   ; L=bank*12
            ld h, 0   ; HL=bank*12
            ; all the banks are 48K except bank 0 which is 64K, adjust for that
            cp 0      ; A=L already
            jr z, bankmmu
            ; we're in banks 1 or above
            add 4
            ld l, a   ; adjust upwards
bankmmu:    ; now we program the MMU!
            ld b, BANKED_PAGES
            ld c, 0
nextpage:   ld a, c
            out (MMU_PAGESEL), a
            ld a, h
            out (MMU_FRAMEHI), a
            ld a, l
            out (MMU_FRAMELO), a
            inc c
            inc hl
            djnz nextpage
            ld a, (activebank)
            or a
            jnz novwin
            ; Video Window
            ; 7000 - 7FFF in Bank 0 is remapped to the video RAM
	    ld a, 7
	    out (MMU_PAGESEL), a
            ld a, 0x20
            out (MMU_FRAMEHI), a
            ld a, 0x10
            out (MMU_FRAMELO), a
novwin:
	    pop bc
	    pop de
noswitch:   pop hl
            ret

activebank: db 255
	    
setbnk:     ; set the bank to be used for the next read/write sector
            ; BDOS calls SETDMA, SETBNK (so we can pick the MMU setup based
            ; on the address)
	    ld (curbank), a
            ret

setdma:     ; set DMA address given by BC
	    ; In CP/M 3 this can be > 128 bytes long
            ld (curdmaaddr), bc
            ret

read:       ld hl, (diskptr)
diskop:     dec hl
            ld d, (hl)
            dec hl
            ld e, (hl)
            ex de, hl
            jp (hl)

write:      ld hl, (diskptr)
            dec hl
            dec hl
            jr diskop

;
;  Flash drive operations. Read only device.
;  We do everything we can in the banked helper routine except the final
;  copy and bottom wipe.
;
flash_read: call flash_read_banked
            ld a, (curbank)     ; We must be in the right bank for the copy
            call selmem
            ld b, 128
flash_xfer: out (SPI_TX), a     ; Dummy write
            in a, (SPI_RX)
            ld (hl), a
            inc hl
            djnz flash_xfer
            ld a, 0xff
            out (SPI_CHIPSELECT), a
            jr rw_done

no_write:   ld a, 2		; Read only
            ret

;
;   SD Card driver, not yet all plumbed in. We push everything but the
;   transfer itself into the banked helper
;
sd_read256:  ld a, 0xff
             out (SD_TX), a
             ini
             jr nz,sd_read256
             ret
sd_read:
	     call sd_read_banked
             ret nz
             ld a, (curbank)
             call selmem
             call sd_read256        ; on exit b = 0
             call sd_read256
             jp rw_done

sd_write:
	     call sd_write_banked
             ret nz
             ld a, (curbank)
             call selmem
sd_write1:   outi
             jr nz, sd_write1
sd_write2:   outi
             jr nz, sd_write2             
             jp rw_done

;
;   Ramdisc driver.
;   This mostly relies upon a banked helper to set up the MMU I/O access
;   and we then inir/otir the block in common so we can hit the correct
;   bank.
;
read_rd
            call ram_prepare_banked
	    ld a, (curbank)
	    call selmem
            inir
rw_done:    xor a ; return to bank 0
	    call selmem
            xor a ; A=0: success
            ret

write_rd    ; write to our RAM disk
            ; In CP/M 3 this is physical sectors
	    ; C is now a re-blocking hint 0 - defer, 1 non-defer, 2 -
            ; deferred write of first sector of new block
            call ram_prepare_banked
	    ld a, (curbank)
	    call selmem
            otir
            jr rw_done

time:       ; FIXME: copy TOD clock into BOOT-0x0C
            ; Must preseve HL, DE. If C=0xFF set time
            ret

	    ; The IRQ handler must in common we could be in any bank when
	    ; this is invoked
clockint:   ; Ripple the clock, once a second, only using AF
	    push af
	    in a, (TIMER_STATUS)
	    bit 7, a
	    jr nz, clockevent
	    pop af
	    ei
	    ret
clockevent: 
	    ld a, (SEC)
	    inc a
	    daa
	    cp 0x60
	    jr nc, clockmin
	    ld (SEC), a
clockret:   xor a
	    out (TIMER_COMMAND), a	; Ack
	    pop af
	    ei
	    ret
clockmin:   xor a
	    ld (SEC), a
	    ld a, (MIN)
	    inc a
	    daa
	    cp 0x60
	    jr nc, clockhr
	    ld (MIN), a
	    jr clockret
clockhr:    xor a
	    ld (MIN), a
	    ld a, (HOUR)
	    inc a
	    daa
	    cp 0x24
	    jr nc, clockday
	    ld (HOUR), a
	    jr clockret
clockday:   xor a
	    ld (HOUR), a
	    ld a, (DATE)
	    inc a
	    ld (DATE), a	; we can store anyway if wrapped
	    jr z, clockday256
	    jr clockret
clockday256:ld a, (DATE+1)
	    inc a
	    ld (DATE+1), a
	    jr clockret

userf:      ; Extended BIOS call
	    ret

;---------------------------------------------------------------------------------------------------------------
; debug functions (ideally to be removed in final version, if we ever get that far!)
strout:     ; print string pointed to by HL
            ld a, (hl)
            cp 0
            ret z
            ld c, a
	    push hl
            call conout
            pop hl
            inc hl
            jr strout

outnibble:  and 0x0f
            cp 10
            jr c, numeral
            add 0x07
numeral:    add 0x30
            ld c, a
            call conout
            ret

outcharhex: ; print byte in A as two-char hex value
            push hl
            push de
            push bc
            push af
            ld d, a ; copy
            rra
            rra
            rra
            rra
            push de
            call outnibble
            pop de
            ld a, d
            call outnibble
            pop af
            pop bc
            pop de
            pop hl
            ret

;---------------------------------------------------------------------------------------------------------------
; scratch RAM used by BIOS
curbank:    db 0
curdisk:    db 0x55
curtrack:   db 0x55
cursector:  dw 0x5555   ; 16bit in CP/M 3
curdmaaddr: dw 0x5555
mmutemp0:   db 0x55
mmutemp1:   db 0x55
mmutemp2:   db 0x55
diskptr:    dw 0        ; current disk dpb entry
;xmove_sec:  dw 0xffff
;xmove_buf:  ds 0x80	; FIXME - check if we can use the BDOS SCB provided
		        ; scratch ?

; Console - video mode must not be banked

VID_MODE:   db 0x00
	    db "</bios>"
; we point the stack at 0x0000 on restart, so that may overwrite the very end of RAM.

;
;	GENCPM places cseg into the common area and dseg into bank 0
;
;	Bank 0 is allowed to contain certain types of driver code with
; 	strict limitations (see CP/M 3 system guide). At the moment we just
;	bank the boot message to show willing. If SD is added we should
;	probably bank most of the SD card drivers.
;
	dseg

bootmsg:    db 10, 13
	    db ' SocZ80 CP/M 3.x Banked BIOS 0.2D'
	    db 10,13
	    db ' Will Sowerbutts, Alan Cox 2014-06-28'
	    db 10,13,0

clockinit:
	    ; 0x000F4240 - 1Hz clock, we don't need more
	    ld a, TIMCMD_SEL_DOWNRESET
	    out (TIMER_COMMAND), a
	    ld a, 0x40 
	    out (TIMER_VAL0), a
	    ld a, 0x42
	    out (TIMER_VAL1), a
	    ld a, 0x0F
	    out (TIMER_VAL2), a
	    xor a
	    out (TIMER_VAL3), a
	    ld a, TIMCMD_DOWNRESET
	    out (TIMER_COMMAND), a
	    in a, (TIMER_STATUS)
	    set 6, a
	    res 7, a
	    out (TIMER_STATUS), a
	    ret
;
; Switch bits
;
switchinit: in a, (SWITCHES)
            bit 2, a	     ; Switch 2: turns on C and P
            jr nz, has_drive3
            ld hl, 0
            ld (ram3onoff), hl
            ld (flash3onoff), hl
has_drive3: ld hl, ccpfcb     ; Boot drive
            bit 3, a
            jr z, ramboot
            ld (hl), 14       ; Boot from flash drive
ramboot:    bit 4, a	      ; Switch 4
            ret z
            inc (hl)          ; Boot from B or O not A or N.
            ret
            
;
; SPI helpers
;
spi_init:
            ld a, 1
            out (SPI_DIVISOR), a    ; 32MHz
            ld a, 0xfe
            out (SPI_CHIPSELECT), a
            ld a, 0x9f 
            out (SPI_TX), a   ; Read ID

            xor a
            out (SPI_TX), a   ; Address 0
            out (SPI_TX), a
            out (SPI_TX), a

            out (SPI_TX), a
            in a, (SPI_RX);
            ld h, a
            out (SPI_TX), a
            in a, (SPI_RX)
            ld l, a
            out (SPI_TX), a
            in a, (SPI_RX)
            ld e, a
            ld a, 0xff
            out (SPI_CHIPSELECT), a
            ld a, h
            cp FLASH_MAN
            jr nz, spi_badflash
            ld a, l
            cp FLASH_DEV1
            jr nz, spi_badflash
            ld a, e
            cp FLASH_DEV2
            ret z
spi_badflash:
	    ld hl, badflash
            call strout
            ; Disable the flash drives
            ld hl, 0
            ld (dphflash), hl
            ld (dphflash + 2), hl
            ld (dphflash + 4), hl
            ret

badflash:   db 'Unknown flash device', 13, 10, 0

;
; SD card driver

SD_CMD0	    equ           0x40
SD_CMD1	    equ	          0x40 + 1
SD_CMD8     equ	          0x40 + 8 
SD_CMD9     equ           0x40 + 9
SD_CMD12    equ           0x40 + 12
SD_CMD16    equ           0x40 + 16
SD_CMD17    equ           0x40 + 17
SD_CMD24    equ           0x40 + 24
SD_CMD55    equ           0x40 + 55
SD_CMD58    equ           0x40 + 58

SD_ACMD41   equ           0xc0 + 41

CT_NONE     equ           0
CT_MMC      equ		  1
CT_SD1      equ           2
CT_SD2      equ           3
CT_BLOCK_SD2 equ	  4

;
;	SD card initialiser
;
sd_init:
           in a, (0x80)
           or a                        ; wings 0, Wills SD card
           ret nz
;
;	Detect the SD card
;
sd_spi_init:
            ld a, 0xff
            out (SD_DIVISOR), a        ; Low speed for detect
            dec a
            out (SD_CHIPSELECT), a     ; 0xfe - chip select

            ld b, 20		       ; now wait 160 clocks
sd_spi_init_pre:
            ld a, 0xFF
            out (SD_TX), a
            in a, (SD_RX)
            djnz sd_spi_init_pre

            ld b, SD_CMD0              ; probe the device
            call sd_cmd_nodata
            cp 1
            jr nz, sd_spi_fail         ; Nothing sane
            ld b, SD_CMD8              ; Get device info
            ld hl, 0x01AA
            call sd_cmd_data16
            cp 1
            jr nz, sd_notsdhc          ; Something but not SDHC
            call sd_cdinfo4            ; HL now points past the data
            dec hl
            ld a, (hl)                 ; check if we are 3.3v
            cp 0xaa
            jr nz, sd_spi_fail         ; Not 3.3v capable
            dec hl
            ld a, (hl)
            cp 0x01
            jr nz, sd_spi_fail         ; Not 3.3v capable
            ld hl, 0
            ld de, 0x8000
            ; We need a few timeouts around here!
sd_spi_idleout:
            ld b, SD_ACMD41            ; Leave idle, HCS
            call sd_cmd
            jr nz, sd_spi_idleout
            ld b, SD_CMD58             ; nearly done
            call sd_cmd_nodata
            jr nz, sd_spi_fail
            ; Success - we have an SDHC card
            call sd_cdinfo4
            ld a, 2
            out (SD_DIVISOR), a
            ld c, CT_SD2
            ld a, (sd_cdinfobuf)
            bit 6, a                   ; check type
            jr z, sd_spi_done
            ld c, CT_BLOCK_SD2
            jr sd_spi_done
sd_spi_fail:ld c, 0                    ; 0 is our error return
sd_spi_done:ld a, 0xff
            out (SD_CHIPSELECT), a     ; deselect the chip
            out (SD_TX), a
            in a, (SD_RX)              ; mop up
            ld a, c                    ; return code into a
            cp CT_MMC
            ld a, 2
            jr nz, sd_clock2
            inc a                      ; SD is slower
sd_clock2:  out (SD_DIVISOR), a
            ld a, c
            ld (sd_cardtype), a
            or a
            ret

sd_notsdhc: ; SDSC or MMC ?
            ld b, SD_ACMD41
            call sd_cmd_nodata
            cp 2
            jr c, sd_mmc3              ; SD or MMC ?
sd_sd1_idle:ld b, SD_ACMD41            ; wait for ready
            call sd_cmd_nodata
            jr nz, sd_sd1_idle
            ld c, CT_SD1               ; set type
sd_blocklen:
            push bc                    ; save return code
            ld b, SD_CMD16
            ld hl, 512
            call sd_cmd_data16
            pop bc
            jr nz, sd_spi_fail
            jr sd_spi_done             ; successful detection

sd_mmc3:    ld b, SD_CMD1              ; MMC card
            call sd_cmd_nodata
            jr nz, sd_mmc3             ; wait for the card
            ld c, CT_MMC
            jr sd_blocklen

;
;	Get size data, return != 0 on error
;
sd_size:   ld b, SD_CMD9
           call sd_cmd_nodata
           ret nz
           ld bc, 16
           call sd_cdinfo
           or a
           ret nz
           ld a, (sd_cdinfobuf)
           bit 6, a
           jr nz, sd_size_sdc2
           ;
           ;   Messy type
           ;
           ;
           xor a 
           ret
sd_size_sdc2:
           ;   Saner block descriptions
           ;  9 + 8 << 8 + 1   Kilosectors
           xor a
           ret

; clock out four bytes into the buffer sd_cdinfobuf
sd_cdinfo4: ld b, 4
; entry point for different sizes (16 max)
sd_cdinfo:  ld hl, sd_cdinfobuf
; entry point for straight copy to HL
sd_spi_d0:  ld a, 0xff                 ; Copy the response
            out (SD_TX), a
            in a, (SD_RX)
            ld (hl), a
            inc hl
            djnz sd_spi_d0
            ret

;
;  Wait for the SD to become ready - receive until we stop getting 0xFF
;  back from the card. On entry c always holds SD_TX
;
sd_wait_ready:
            ; Preserve all but A
	    ld a, 0xff
            out (c), a
            in a, (SD_RX)
sd_wait_ready2:
	    ld a, 0xff
            out (c), a
            in a, (SD_RX)
            cp 0xff
            jr nz, sd_wait_ready2
            ret

;
; Wait to receive a block, on entry C is alway SD_RX
;
sd_wait_rx:
	    ld a, 0xff
            out (SD_TX), a
            in a, (c)
            cp 0xff
            jr z, sd_wait_rx
            xor 0xfe        ; return 0 if 0xfe
            ret z
            ld a, 1         ; CP/M physical error
            ret
            
;
;   Main command processor. This is slightly complicated by the need to
;   juggle a 32 bit argument and a command in registers, as well as the
;   fact that we need to consider that some commands are sent with CMD55
;   first as sort of 'shift key'. We also fake CRC values on a couple of
;   commands as needed by the protocol rather than doing real CRC.
;
;   d,e,h,l holds the argument and b holds the command
;
sd_cmd_nodata:
	    ld hl, 0        ; low 16bits of arg in HL
sd_cmd_data16:
            ld de, 0        ; high bits in de
sd_cmd:     ; command b with 0 arg and not an ACMD 
            bit 7, b
            ; ACMD ?
            jr z, sd_notacmd
            ; Send CMD 55, 0 first
            ; use the alternate register bank as we don't save it for IRQs
            ; or stuff like that
            exx
            ld b, SD_CMD55
            call sd_cmd_nodata
            exx
            res 7, b
sd_notacmd:
            ld a, 0xff               ; deselect
            ld c, SD_TX
            out (SD_CHIPSELECT), a
            dec a                    ; reselect (0xfe)
            out (SD_CHIPSELECT), a
            call sd_wait_ready       ; wait for the card to respond
            out (c), b               ; send the command and data
            out (c), d
            out (c), e
            out (c), h
            out (c), l
            ld a, b                  ; check for the CRC cases
            ld e, 0x01               ; no CRC
            cp SD_CMD0
            jr nz, sd_notcmd0
            ld e, 0x95
            jr sd_crc_tx
sd_notcmd0: cp SD_CMD8
            jr nz, sd_crc_tx
            ld e, 0x87
sd_crc_tx:  out (c), e               ; send the correct fudged CRC
            cp SD_CMD12
            jr nz, sd_noskip         ; might be better to just offset bump read?
            ld a, 0xff
            out (c), a
            in a, (SD_RX)       ; Meh - we should fix the hw so rx and tx
				; are the same offset!
sd_noskip:  ld b, 10
sd_cmd_rx:  ld a, 0xff
            out (c), a
            in a, (SD_RX)
            bit 7, a
            ret z
	    djnz sd_cmd_rx
            or a                ; Many callers want to check v 0
            ret            



;
;   Block offset into d,e,h,l
;
sd_make_arg:
            ld a, (sd_cardtype)
            cp CT_BLOCK_SD2
            jr z, sd_make_block
            ; 16384 * 128 byte records / track
            ; deblocked by CP/M so we see 512 byte size request/offset
            ; ie 4096 blocks / track
            ; 000TTTTT TTTSSSSS SSSSSSS0 00000000
            ld bc, (cursector)
            sla c
            rl b
            ld a, (curtrack)
            ld e, 0
            ld l, e    ; l in the final result is fixed at 0
            rrca
            rr e
            rrca
            rr e
            rrca
            rr e
            and 0x1F
            ld d, a    ; High bits
            ld a, b
            and 0x1f
            or e
            ld e, a    ; bits 24-16
            ld h, c    ; bits 15-8
            ret


sd_make_block:
;
;   Block addressed
;   00000000 0000TTTTTTTTSSSS SSSSSSSS
;
;
            ld a, (curtrack)
            rlca                  ; Swap the nibbles over
            rlca
            rlca
            rlca
            ld d, a               ; Save a copy
            and 0x0f              ; top nibble
            ld e, a               ; save in result e
            ld hl, (cursector)    ; sector forms the low 12 bits
            ld a, d               ; high bits from swapped track nibble
            and 0xf0              ; drop the low bits
            or h                  ; add sector
            ld h, a               ; return in h
            ld d, 0               ; top of address is 0, l was set earlier
            ret			  ; done



sd_read_banked:
;
;	Set up for a read. 
;       On return nz is true if an error (and the code is set)
;                 z is true on no error (and the xfer is ready to begin)
;
             call sd_make_arg
             ld b, SD_CMD17
             call sd_cmd
             ld hl, (curdmaaddr)
             ld bc, SD_RX    ; b=0, c=port
             jp z, sd_wait_rx
sd_hw_err:   ld a, 1
             ret
;
;	Similar for write
;
sd_write_banked:
	     call sd_make_arg
             ld b, SD_CMD24
             call sd_cmd
             ld hl, (curdmaaddr)
             jr nz, sd_hw_err
             ld bc, SD_TX
             call sd_wait_ready
             xor a
             ret

sd_cdinfobuf ds 16
sd_cardtype db 0

;
; Bank as much as we can of the flash reader
;
flash_read_banked:
	    ld a, 0xfe
            out (SPI_CHIPSELECT), a
            ld a, 0x0b
            out (SPI_TX), a     ; fast read
            ld a, (curdisk)
            sub 12              ; 1/2/3 are the flash drive bases
            rrca                ; roll it round into bits 6:5 with 7 as 0
            rrca
            rrca
            ld b, a
            ld a, (curtrack)    ; bits 4-0 are track bits 6-2
            sra a               ; relies on <= 127 sectors/track
            sra a
            or b
            out (SPI_TX), a
            
            ld a, (curtrack)    ; we want the low bits of track but in the top
            rrca                ; move them into the top two bits
            rrca
            and 192             ; remove the chaff
            ld b, a
            ld a, (cursector)
            sra a               ; we know 0-127 so top bit will be 0, low
                                ; bit goes to carry
            or b                ;
            out (SPI_TX), a     ; high 6 bits of sector + bottom 2 of track
            ld a, (cursector)
            rrca		; recovery carry bit
            and 128
            out (SPI_TX), a     ; sector bottom bit
            out (SPI_TX), a     ; Dummy write
            ld hl, (curdmaaddr) ; Where to copy data
            ret

;
; Banked elements of RAM driver. This has some strong similarities to the
; flash driver and it might be nice to extract the maths into a helper
;
ram_prepare_banked:
            ld a, 0xff
            out (MMU_PAGESEL), a
            xor a
            out (MMU_PTR_VAL0), a
            ld a, (curdisk)     ; disk 1-3 maps to 2MB chunk 1 to 3
            inc a
            rrca                ; roll it round into bits 6:5 with 7 as 0
            rrca
            rrca
            ld b, a
            ld a, (curtrack)    ; bits 4-0 are track bits 6-2
            sra a               ; relies on <= 127 sectors/track
            sra a
            or b
            out (MMU_PTR_VAL1), a
            
            ld a, (curtrack)    ; we want the low bits of track but in the top
            rrca                ; move them into the top two bits
            rrca
            and 192             ; remove the chaff
            ld b, a
            ld a, (cursector)
            sra a               ; we know 0-127 so top bit will be 0, low
                                ; bit goes to carry
            or b                ;
            out (MMU_PTR_VAL2), a     ; high 6 bits of sector + bottom 2 of track

            ld a, (cursector)
            rrca		; recovery carry bit
            and 128
            out (MMU_PTR_VAL3), a     ; sector bottom bit

            ld hl, (curdmaaddr) ; Where to copy data
            ld bc, 128 * 256 + MMU_DATA; copy size, port
            ret
;
; Banked parts of the console init
;
coninit:    ;; Console
            in a, (0xDF)    ; Check for video support
            and 0xF8;       ; strip low bits
            or a
            ret nz
            ld a, 0x11
            out (0xde), a   ; text mode, 80 column, ASCII type font
            ld a, 0xF6      ; dull white
            out (0xd8), a   ; ink
            ld a, 0xff      ; white
            out (0xda), a   ; bright ink
            ld a, 1         ; mode - force for now
            ld (VID_MODE), a
            call con_wipe    ; blank console
            ld a, 128 + 32   ; UART 0 + video
            ld (COVEC+1), a
	    ret
;
;  Banked console helpers
;
; Reverse Scroll
;
con_rscroll:  
            ld a, (VID_FLAGS)
            bit 0, a
            jr nz, con_rscroll40
            ; Reverse scroll needs to be done from the bottom backwards
            ld hl, 0x7000 + 30 * 80 - 1  ; last line but one of frame buffer
            ld de, 0x7000 + 29 * 80 - 1  ; line above
            ld bc, 29 * 80     ; ditto
            lddr
            jr llwipe
con_rscroll40:
            ; Reverse scroll needs to be done from the bottom backwards
            ld hl, 0x7000 + 30 * 40 - 1  ; last line but one of frame buffer
            ld de, 0x7000 + 29 * 40 - 1  ; line above
            ld bc, 29 * 40     ; ditto
            lddr
            jr llwipe

;
;  Wipe. Preserve character in C (FIXME: may be able to lose this)
;
con_wipe:   push bc
            ld a, (VID_FLAGS)
            ld bc, 30 * 80 - 1
            bit 0, a
            jr  z, con_wipe80
            ld bc, 30 * 40 - 1
con_wipe80: 
            ld hl, 0x7000
            ld de, 0x7001
            ld (hl), 0
            ldir
            pop bc
            ret
;
; Scroll
;
con_scroll:  
            ld hl, 0x7000 + 80 ; start of frame buffer + 1 line
            ld bc, 29 * 80     ; bytes to copy
            ld a, (VID_FLAGS)
            bit 0, a
            jr z, con_scroll80
            ld hl, 0x7000 + 40 ; ditto for 40 column
            ld bc, 29 * 40
con_scroll80:
            ld de, 0x7000      ; line below
            ldir               ; de now points at the start of the last
llwipe:     ex de,hl           ; line which needs clearing
            ld b, 80
            bit 0, a
            jr z, llset
            sra b              ; 40 column not 80
llset:      ld (hl), 32        ; space
            inc hl
            djnz llset 
	    ret
;
;  Escape states
;
con_escy1:  ld a, 3
            ld (VID_STATE), a
            ld a, c
            ld (VID_NEWX), a
            ret
con_escy2:  xor a
            ld (VID_STATE), a
            ld a, c
            sub 32
            jc con_escy2_x    ; bad range
            cp 30
            jnc con_escy2_x   ; off bottom
            ld (VID_Y), a
            ; VT52 trivia - specifying an invalid Y to move in X only
	    ; is specifically allowed in the manual
con_escy2_x:
            ld a, (VID_NEWX)
            ld hl, VID_COLS
            sub 32
            ret c
            cp (hl)
            ret nc
            ld (VID_X), a
            ret
con_quoted:
            xor a
            ld (VID_STATE), a
            ld a, c
            jr con_normal

;
;  Entry point for banked console
;
;  FIXME: Cursor
;
conentry:   ld a, (VID_STATE)
            or a
            jz con_std        ; not mid special processing
            dec a             ; escape is mode 1
            jz con_escproc    ; escape processor
	    dec a
            jz con_escy1      ; escape Y 1
            dec a
            jz con_escy2      ; escape Y 1 2
            dec a
            jz con_quoted     ; mode 4 is quoted
            ; fall through on an invalid mode
con_std:    ld a, c
            cp 32
            jnc con_normal    ; 32+ are normal codes
            ;; cp 7
            ;; jz con_bell       ; Bell ???
            cp 8              ; 8 is backspace
            jz con_bs
            cp 9
            jz con_tab        ; to next 8 char stop
            cp 10             ; newline
            jz con_lf         ; line feed
            cp 12
            jz conesc_clr     ; page feed
            cp 13
            jz con_cr         ; carriage return
            cp 27
            jz con_esc        ; escape sequences
            ;; lots to add
            ret
con_esc     ld a, 1
            ld (VID_STATE), a
            ret
con_bs      ld a, (VID_X);
            or a
            ret z
            dec a
            ld (VID_X), a
            ret
con_cr:     xor a
            ld (VID_X), a
            ret
con_lf:     ld a, (VID_Y)
            inc a
            cp 30
            jc con_noscr
            call con_scroll
            jr con_leftedge
con_noscr:
            ld (VID_Y), a
con_leftedge:
            xor a
            ld (VID_X), a
            ret
con_tab:    ld c, 32		; tab - space until we hit a tab mark
con_tab2:   call con_normal
            ld a, (VID_X)
            and 7
            jr nz, con_tab2
            ret
con_normal: 			; Normal data
            call con_char       ; Draw the character
            ld a, (VID_X)	; Each byte bumps on the cursor
            inc a	        ; Have we finished ?
            ld (VID_X), a
            ld hl, VID_COLS     ; Column for wrapping
            cp (hl)		; Character pushes us over line end ?
            ret c	        ; If not keep y and go printing
            xor a               ; Wrap to start of line
            ld (VID_X), a
            ld a, (VID_Y)       ; Bump y
            inc a
            cp 30
            jp nc, con_scroll   ; Bottom of screen, cursor stays the same
cony_fit:   ld (VID_Y), a
            ret
;
;  We need a general purpose console output 'print char c at d,e' type fn
;  Also to support 40 or 80 chars
;

; con_addr MUST preserve BC
con_addr:   ld a, (VID_Y)
            ld l, a             ; Character at y + 40/80 * x
            ld h, 0
            add hl, hl
            add hl, hl
            add hl, hl
            push hl
            add hl, hl
            add hl, hl
	    pop de
 	    add hl, de   ; A * 40
            ld a, (VID_FLAGS)
            bit 0,a     ; 40 colum ?
            jr nz, con_addr_40
            add hl, hl   ; A * 80
con_addr_40:
            ld a, (VID_X)
            ld e, a     ; X column
            ld d, 0x70  ; And do the bus address while we are at it
            add hl, de; Correct target square
            ret

con_char:   call con_addr
            ld (hl), c
            ret

; Entered with C = symbol
con_escproc:
            ld a, c
            cp 'A'
            jz conesc_up
            cp 'B'
            jz conesc_down
            cp 'C'
            jz conesc_right
            cp 'D'
            jz conesc_left
            cp 'E'
            jz conesc_clr
            cp 'H'
            jz conesc_home
            cp 'I'
            jz conesc_rlf   ; cursor up, reverse line feed if at top
            cp 'J'
            jz conesc_clend
            cp 'K'
            jz conesc_cleol
            cp 'L'
            jz conesc_insl
            cp 'M'
            jz conesc_dell
            cp 'Y'
            jz conesc_move
            cp 'Z'
            jz conesc_ident  ; reply ESC / K for VT52
            ; The following probably make no sense to add
            ; '[' - hold screen mode, \ exit hold screen mode
            ; '=' - alternate keypad, ')' exit alternate keypad
            ; 'F' - graphics mode 'G' - exit graphics char mode
            ; We still need to add more of our own for
            ; set ink, paper, bright (generalise it for colour setting
            ; on many palette sizes)
            cp '4'
            jz conesc_40
            ; 80 Column
            cp '8'
            jz conesc_80
            ; Literal symbol
            cp '"'
            jz conesc_quote
            ; Standard font
            cp '0'
            jz conesc_f0
            ; Alternate font
            cp '1'
            jz conesc_f1
            cp '#'
            jz conesc_tton
            cp '@'
            jz conesc_ttoff
            cp 'i'
            jz conesc_inverse
            cp 'b'
            jz conesc_bright
            cp '?'
            jz conesc_reveal
            cp 'h'
            jz conesc_hide
con_escret: xor a
            ld (VID_STATE),a
            ret
conesc_up:  ld a, (VID_Y)
            dec a
            jr c, con_escret
            ld (VID_Y), a
            jr con_escret
conesc_down:
            ld a, (VID_Y)
            inc a
            cp 30
            jr nc, con_escret
            ld (VID_Y), a
            jr con_escret
conesc_right:
	    ld a, (VID_X)
            inc a
            ld hl, VID_COLS
            cp (hl)
            jr nc, con_escret
            ld (VID_X), a
            jr con_escret
conesc_left:
            ld a, (VID_X)
            dec a
            jr c, con_escret
            ld (VID_X), a
            jr con_escret
conesc_40:
            ld a, 0x10
            out (0xDE), a
            ld hl, VID_FLAGS
            set 0, (hl)
            inc hl
            ld (hl), 40
            jr conesc_clr
conesc_80:
            ld a, 0x11
            out (0xDE), a
            ld hl, VID_FLAGS
            res 0, (hl)
            inc hl
            ld (hl), 80
            ; Fall through
conesc_clr:
	    call con_wipe
conesc_home:
	    xor a
            ld (VID_X), a
            ld (VID_Y), a
            jr con_escret
conesc_move:
	    ld a, 2
            ld (VID_STATE), a
            ret
conesc_rlf:
            ld a, (VID_Y);
            or a
            jr nz, conesc_up
            call con_rscroll
            jr con_escret
conesc_f0:  in a, (0xDE)
            and 0xF7
conesc_f0_put:
            out (0xDE), a
            jr con_escret
conesc_f1:  in a, (0xDE)
            or 0x10
            jr conesc_f0_put
conesc_tton: in a, (0xDE)
            or 0x04
            jr conesc_f0_put
conesc_ttoff:
            in a, (0xDE)
            and 0xFB
            jr conesc_f0_put
conesc_inverse:
            in a, (0xDE)
            or 0x02
            jr conesc_f0_put
conesc_bright:
            in a, (0xDE)
            and 0xFE
            jr conesc_f0_put
conesc_reveal:
            in a, (0xDE)
            or 0x08
            jr conesc_f0_put
conesc_hide:
            in a, (0xDE)
            and 0xF7
            jr conesc_f0_put
conesc_cleol:
            ; FIXME: slow way to do it, but probably fine for 80 chars!
            ld hl, VID_X
            ld a, (VID_COLS)
            ld c, (hl)
            sub c            ; A is now chars to wipe, we save C for later
            ld b, a
conesc_cllp:push bc
            push hl
            ld c, 32
            call con_char
            pop hl
            pop bc
            inc (hl)         ; Next column
            djnz conesc_cllp
            ; Restore cursor position
            ld (hl), c
            jp con_escret

conesc_clend:
            call con_addr    ; Current address
            ld de, 0x7000 + 40 * 30 ; Byte after end
            ld a, (VID_FLAGS)
            bit 0, a
            jr nz, conesc_clend_40
            ld de, 0x7000 + 80 * 30
conesc_clend_40:
            ex de, hl        ; We want to compute end - current 
            ccf
            sbc hl, de       ; DE is our address, HL is now the length
            ex de, hl        ; HL back to write address, DE to length
            ld b, e          ; do the stragglers first
            inc d
conesc_clrit:
            ld (hl), 32
            inc hl
            djnz conesc_clrit
            dec d
            jr nz, conesc_clrit
            jp con_escret
            
                        
; These two are in a lot of VT52-alikes but not real VT52
conesc_insl:
conesc_dell:
; Not clear what this ought to do
conesc_ident:
            jp con_escret
conesc_quote:
; Next byte is copied literally
            ld a, 4
            ld (VID_STATE), a
            ret


; Console
VID_Y:      db 0x00
VID_X:      db 0x00
VID_NEWX:   db 0x00
VID_STATE:  db 0x00
VID_FLAGS:  db 0x00  ; cols must directly follow flags
VID_COLS:   db 80    ; column we wrap at (not the same as display width)
